#!/usr/bin/env python2
# -*- coding=utf-8 -*-
'''
    search pubmeds quickly
'''
import os
import re
import sys
import json
import time
import datetime
import argparse
import textwrap
import codecs

import bs4
import requests
import colorama
import django
import xlwt
from django.conf import settings
from django.template import loader
from googletrans import Translator


BASE_DIR = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(0, os.path.dirname(BASE_DIR))


from pubmed2.tools import GetIF, try_again

from pubmed2.info import __version__, __author__, __author_email__


reload(sys)
sys.setdefaultencoding('utf8')

colorama.init()

color_dict = {
    'fore_reset': colorama.Fore.RESET,
    'back_reset': colorama.Back.RESET,
    'fore_red': colorama.Fore.RED,
    'fore_green': colorama.Fore.GREEN,
    'fore_cyan': colorama.Fore.CYAN,
    'fore_yellow': colorama.Fore.YELLOW,
    'back_yellow': colorama.Back.YELLOW,
    'style_bright': colorama.Style.BRIGHT,
}


def configure_django():

    settings.configure(
        DEBUG=True,
        TEMPLATE_DEBUG=True,
        TEMPLATES=[{
            'BACKEND':
            'django.template.backends.django.DjangoTemplates',
            'DIRS': [BASE_DIR, os.path.join(BASE_DIR, 'templates')]
        }])
    django.setup()


def get_now_time(time_fmt='%Y-%m-%d %H:%M:%S'):

    return datetime.datetime.now().strftime(time_fmt)


class Pubmed(object):

    def __init__(self, **args):

        self.args = args
        self.term = args.get('term')
        self.retmax = args.get('retmax')
        self.format = args.get('out_format')
        self.min_factor = args['min_impact_factor']
        self.out = args.get('out_prefix') or \
            re.sub(r' |\(|\)|/', '_', self.term.strip()).strip('_')

        self.each_page_max = args['page_size']
        self.encoding = args['encoding']
        self.start_point = args['start_point']

        self.not_trans = args['not_translate']

        self.title = args['title'].split(',')

        self.translator = Translator(service_urls=['translate.google.cn'])
        self.BASE_URL = 'https://eutils.ncbi.nlm.nih.gov/entrez/eutils/'
        self.timeout = 60

    def start(self):

        for field in self.title:
            if field not in default_title:
                print '[error] invalid title field "{}"\nyou can only choose from {}'.format(field, default_title)
                exit(1)

        if self.not_trans and 'abstract_cn' in self.title:
            self.title.pop(self.title.index('abstract_cn'))

        if self.min_factor == -1 and 'impact_factor' in self.title:
            self.title.pop(self.title.index('impact_factor'))

        # print 'output title: {}'.format(self.title)

        # term_list = self.term.split(',')
        term_list = re.split(r',|\s+|;', self.term)

        total_length = ''
        if self.term.isdigit():
            pmids = [self.term]
        elif all([each.isdigit() for each in term_list]):
            pmids = term_list
            self.out = '{}_{}'.format(pmids[0], pmids[-1])
        elif os.path.isfile(self.term):
            with open(self.term) as f:
                pmids = [line.strip() for line in f]
        else:
            total_pmids = self.get_pmids(self.term)
            if self.retmax == 0 or len(total_pmids) <= self.retmax:
                pmids = total_pmids
            else:
                pmids = total_pmids[:self.retmax]
                total_length = '/{}'.format(len(total_pmids))

        if not pmids:
            print '{fore_red}[error {time}] No pubmed for term: "{term}"{fore_reset}'.format(
                time=get_now_time(), term=self.term, **color_dict)
            exit(0)

        print '{fore_red}{back_yellow}Use {length}{total_length} pmids:{back_reset}{fore_reset} {pmids}'.format(
            length=len(pmids),
            total_length=total_length,
            pmids=pmids
            if len(pmids) <= 10 else '[{}, ...]'.format(', '.join(pmids[:10])),
            **color_dict)

        if self.start_point < len(pmids):
            print 'start craw from the {}st pmid'.format(self.start_point)
            pmids = pmids[self.start_point - 1:]

        xmls = self.get_xmls(pmids)

        # results = self.parse_xml(xmls, pmids)
        # results = [result for result in self.parse_xml(xmls, pmids)]
        results = self.parse_xml(xmls, pmids)

        result_len = self.save_result(results)

        if self.min_factor:
            print '{fore_yellow}\nNumber of results(IF>={mif}): {number}{fore_reset}'.format(
                time=get_now_time(),
                mif=self.min_factor,
                number=result_len,
                **color_dict)

    def save_result(self, results):

        # write from generator
        if self.format == 'xls':
            with codecs.open(
                    self.out + '.xls',
                    'w',
                    encoding=self.encoding,
                    errors='ignore') as out:
                out.write('\t'.join(self.title) + '\n')
                for num, result in enumerate(results):
                    linelist = [result[each] for each in self.title]
                    out.write('\t'.join(linelist) + '\n')
            print 'save result to: {}.xls'.format(self.out)
            return num + 1

        # otherwise, convert generator to list
        results = list(results)

        if self.format in ('xls', 'all'):
            with codecs.open(
                    self.out + '.xls',
                    'w',
                    encoding=self.encoding,
                    errors='ignore') as out:
                out.write('\t'.join(self.title) + '\n')
                for result in results:
                    linelist = [result[each] for each in self.title]
                    out.write('\t'.join(linelist) + '\n')
            print 'save result to: {}.xls'.format(self.out)

        if self.format in ('xlsx', 'all'):
            wb = xlwt.Workbook()
            ws = wb.add_sheet('result')


            # align: wrap yes,vert centre, horiz left;
            # font: name Arial, bold True, height 200;
            # pattern: pattern solid,fore-colour light_yellow;

            header_style = xlwt.easyxf('''
                font: name Times New Roman, bold on, height 200;
                pattern: pattern solid, fore-colour white;
            ''')

            row_styles = [
                xlwt.easyxf('pattern: pattern solid, fore-colour light_green;'),
                xlwt.easyxf('pattern: pattern solid, fore-colour light_yellow;')
            ]

            row = 0
            col = 0
            for name in self.title:
                ws.write(row, col, name.upper(), header_style)
                col += 1
            row += 1

            for each in results:
                col = 0
                for name in self.title:
                    value = each[name]
                    if name == 'impact_factor' and value != '.':
                        value = float(value)
                    ws.write(row, col, value, row_styles[row % 2])
                    col += 1
                row += 1

            if 'abstract_cn' in self.title:
                ws.col(self.title.index('abstract_cn')).width = 256 * 50

            wb.save(self.out + '.xls')

            print 'save result to: {}.xls'.format(self.out)

        if self.format in ('json', 'all'):
            with open(self.out + '.json', 'w') as out:
                # with codecs.open(self.out + '.json', 'w', encoding='utf8', errors='ignore') as out:
                json.dump(results, out, indent=2, ensure_ascii=False)
            print 'save result to: {}.json'.format(self.out)

        if self.format in ('html', 'all'):
            configure_django()
            template = loader.get_template('index.html')
            html = template.render({'results': results})
            # html = template.render({'results': json.dumps(results).replace("'", "\\'")})

            with codecs.open(
                    self.out + '.html', 'w', encoding='utf8',
                    errors='ignore') as out:
                out.write(html)

            print 'save result to: {}.html'.format(self.out)

        return len(results)

    @try_again()
    def get_response(self, url, **kwargs):

        return requests.get(url, **kwargs)


    @try_again()
    def get_pmids(self, term):

        url = self.BASE_URL + 'esearch.fcgi'

        payload = {
            'db': 'pubmed',
            'retmode': 'json',
            'term': term,
            'retmax': '65535'
        }

        return self.get_response(
            url, params=payload,
            timeout=self.timeout).json()['esearchresult']['idlist']

    @try_again()
    def get_xmls(self, pmids):

        url = self.BASE_URL + 'efetch.fcgi'

        for i in range(0, len(pmids), self.each_page_max):
            pmid_list = pmids[i:i + self.each_page_max]

            if i + self.each_page_max >= len(pmids):
                end = len(pmids)
            else:
                end = i + self.each_page_max
            print '{fore_green}===== dealing with pmids: {start} ~ {end}{fore_reset}'.format(
                start=i + 1, end=end, **color_dict)

            payload = {'db': 'pubmed', 'rettype': 'abstract', 'id': pmid_list}

            xml = self.get_response(url, params=payload, timeout=self.timeout).text
            yield xml

    @try_again(10)
    def translate(self, text):

        return self.translator.translate(text, dest='zh-cn').text

    @try_again()
    def parse_xml(self, xmls, pmids):

        # results = []

        n = 1
        for xml in xmls:
            try:
                soup = bs4.BeautifulSoup(xml, 'lxml')
            except Exception:
                print '[warn] lxml was not installed, use html.parser instead'
                soup = bs4.BeautifulSoup(xml, 'html.parser')

            for pubmedarticle in soup.select('pubmedarticle'):
                pmid = self.get_text(pubmedarticle, 'pmid')
                print '{fore_cyan}[info {time}] {n}/{length} dealing with pmid: {pmid}{fore_reset}'.format(
                    time=get_now_time(),
                    n=n,
                    length=len(pmids),
                    pmid=pmid,
                    **color_dict)
                n += 1
                pubdate = self.get_text(pubmedarticle, 'pubdate')
                title = self.get_text(pubmedarticle, 'articletitle')
                abstract = self.get_text(pubmedarticle, 'abstracttext')

                if 'abstract_cn' in self.title:
                    abstract_cn = self.translate(abstract)

                pmc = self.get_text(pubmedarticle, 'articleid[idtype="pmc"]')
                doi = self.get_text(pubmedarticle, 'articleid[idtype="doi"]')
                journal = self.get_text(pubmedarticle, 'journal isoabbreviation')
                journal_full = self.get_text(pubmedarticle, 'journal title')
                issn = self.get_text(pubmedarticle, 'journal issn')
                author_lastnames = pubmedarticle.select(
                    'authorlist author lastname')
                author_initials = pubmedarticle.select(
                    'authorlist author initials')
                authors = [
                    ' '.join([each[0].text, each[1].text])
                    for each in zip(author_lastnames, author_initials)
                ]
                authors = ', '.join(authors)

                pubtype = pubmedarticle.select(
                    'publicationtypelist publicationtype')
                pubtype = '|'.join([each.text for each in pubtype])

                if 'impact_factor' in self.title:
                    impact_factor = GetIF().get_impact_factor(searchname=journal, database=self.args['database']) or '.'

                    if self.min_factor:
                        if (impact_factor == '.') or (float(impact_factor) < self.min_factor):
                            continue

                context = {}
                for each in self.title:
                    context.update({each: locals()[each]})

                yield context
                # results.append(context)

        print '{fore_green}[info {time}] all pmids done!{fore_reset}\n'.format(
            time=get_now_time(), **color_dict)
        # return results

    @staticmethod
    def get_text(soup, key):

        if soup.select(key):
            text = soup.select(key)[0].text.strip().replace('\n', '-')
            return text
        return '.'


def main():

    global default_title
    default_title = ['pmid', 'title', 'pubdate', 'authors', 'abstract', 'abstract_cn', 'journal', 'impact_factor', 'pmc', 'doi', 'pubtype', 'issn']

    epilog = '''
    example: \033[36mpython pubmed.py 'ngs AND disease'
             python pubmed.py '(LVNC) AND (mutation OR variation)' -m 50 -mif 5\033[0m

    contact: {} <{}>
    '''.format(__author__, __author_email__)

    parser = argparse.ArgumentParser(
        prog='pubmed',
        version=__version__,
        formatter_class=argparse.RawTextHelpFormatter,
        epilog=textwrap.dedent(epilog))

    parser.add_argument(
        'term',
        help=(
            '{fore_green}The input term to search\n'
            'could be like NCBI search format. eg: "(NGS) AND CVD"\n'
            'or pmid. eg: 25261758\n'
            'or pmids. eg: 25261758,24564649,24191723\n'
            'or a file contains pmids, one pmid per line{fore_reset}').format(
                **color_dict))
    parser.add_argument(
        '-o', '--out-prefix', help='The prefix of the output filename')
    parser.add_argument(
        '-O',
        '--out-format',
        help='The output format[default=%(default)s]',
        default='xlsx',
        choices=['xls', 'xlsx', 'html', 'json', 'all'])
    parser.add_argument(
        '-m',
        '--retmax',
        type=int,
        help='The max count to return[default=%(default)s], 0 for no limits',
        default=100)
    parser.add_argument(
        '-p',
        '--page-size',
        type=int,
        help='The max number of each page to craw[default=%(default)s]',
        default=100)
    parser.add_argument(
        '-enc',
        '--encoding',
        help='The encoding of output xls file[default="%(default)s"]',
        default='gbk',
        choices=['gbk', 'utf8', 'utf-8'])
    parser.add_argument(
        '-mif',
        '--min-impact-factor',
        type=float,
        help='The minimum of impact factor to save, -1 for no impact_factor')
    parser.add_argument(
        '-start',
        '--start-point',
        type=int,
        help='The start point for all pmids',
        default=1)
    parser.add_argument(
        '-nt',
        '--not-translate',
        action='store_true',
        help='Do not translate abstract')
    parser.add_argument(
        '-t',
        '--title',
        help='The title to craw\n'
        'you can choose one or more from [%(default)s]\n'
        'and separate by ","',
        default=','.join(default_title))
    parser.add_argument(
        '-d',
        '--database',
        help='the impact factor database[default=%(default)s]',
        default=os.path.join(BASE_DIR, 'tools', 'impact_factor.sqlite3'))

    args = vars(parser.parse_args())

    print 'searching term: "{}"'.format(args['term'])

    start = time.time()

    pubmed = Pubmed(**args)
    pubmed.start()

    end = time.time()

    print '\nTotal time: {:.2f}s'.format(end - start)


if __name__ == "__main__":

    main()


#!/usr/bin/env python2
# -*- coding=utf-8 -*-
import os
import sys
import re
import argparse

import fuzzywuzzy.process

from pubmed2.tools import utils


__author__ = 'suqingdong'
__version__ = '3.0'

BASE_DIR = os.path.dirname(os.path.abspath(__file__))
BASE_URL = 'http://www.letpub.com.cn/index.php'

reload(sys)
sys.setdefaultencoding('utf8')


class GetIF(object):

    def __init__(self, **kwargs):

        self.__dict__.update(kwargs)

        self.logger = utils.logger()

    def get_impact_factor(self, searchname=None, database=None):
        
        searchname = searchname or ' '.join(self.__dict__['searchname']).replace('.', '').strip()
        database = database or self.__dict__['database']

        searchname = re.sub(r'\(.*?\)|\.', '', searchname).strip()

        try:
            import sqlite3
            sqlite_not_installed = False
        except ImportError as e:
            print '[warn] SQLite not installed...'
            sqlite_not_installed = True

        if not os.path.exists(database) or sqlite_not_installed:
            print 'get impact factor from website ...'
            payload = {
                'page': 'journalapp',
                'view': 'search',
                'searchsort': 'relevance',
                'searchname': searchname
            }

            result = {}
            for journal, impact_factor, journal_full, _ in self.craw_impact_factor(payload):
                result[journal] = impact_factor
                result[journal_full] = impact_factor
        else:
            print 'search for: {searchname}'.format(**locals())

            # cmd  = '''
            #     sqlite3 {database} 'SELECT * FROM factor WHERE journal LIKE "%{searchname}%";'
            # '''.format(**locals()).strip()
            with sqlite3.connect(database) as conn:
                conn.text_factory = bytes
                cursor = conn.cursor()
                sql = 'SELECT * FROM factor WHERE journal LIKE "%{searchname}%";'.format(**locals())
                # print sql
                cursor.execute(sql)

                result = cursor.fetchall()
                # print result

            if not result:
                print '[warn] no journal names "{searchname}"'.format(**locals())
                return

            result = dict(result)

        match_journal, match_score = fuzzywuzzy.process.extractOne(searchname, result.keys())
        match_impact_factor = result[match_journal]

        if match_score < 90:
            info = 'bad match: {match_journal} [IF={match_impact_factor}] (Sore={match_score})'
        else:
            info = 'best match: \033[32m{match_journal}\033[0m [IF=\033[32m{match_impact_factor}\033[0m]] (Sore={match_score})'
            
        print info.format(**locals())

	return match_impact_factor

    @utils.try_again(5)
    def craw_impact_factor(self, payload):

        soup = utils.get_soup(BASE_URL, params=payload)

        table_index = 0 if 'searchname' in payload else 1

        table = soup.select('table.table_yjfx')[table_index]

        for tr in table.select('tr')[2:-1]:
            tds = tr.select('td')

            if len(tds) < 3:
                yield '', '.', '', ''

            journal = tds[1].contents[-1].text
            journal_full = tds[1].contents[-4].text
            style = tds[1].contents[0].attrs['style']
            impact_factor = tds[2].text

            issn = tds[0].text

            if 'text-decoration:underline' not in style:
                continue

            yield journal, impact_factor, journal_full, issn

    @utils.try_again(5)
    def get_max_page(self):

        payload = {'page': 'journalapp'}

        soup = utils.get_soup(BASE_URL, params=payload)

        max_page = re.findall(
            r'/(\d+)',
            soup.select('form[name="jumppageform"]')[0].text)

        max_page = int(max_page[0])

        self.logger.info('Total page: {max_page}'.format(**locals()))

        return max_page


    def update_database(self):

        payload = {
            'page': 'journalapp',
            'fieldtag': '',
            'firstletter': '',
            'currentpage': 1
        }

        max_page = self.get_max_page()

        out_xls = os.path.splitext(self.__dict__['database'])[0] + '.xls'

        journal_done = {}

        with open(out_xls, 'w') as out:

            while True:
                self.logger.info('\033[31mDealing with page {currentpage}/{max_page}\033[0m'.format(max_page=max_page, **payload))

                for journal, impact_factor, journal_full, issn in self.craw_impact_factor(payload):
                    if journal.strip():

                        # 去重
                        if journal_done.get(journal):
                            continue
                        journal_done[journal] = True

                        self.logger.info('Journal: {journal}({journal_full})\tIF: {impact_factor}'.format(**locals()))

                        line = '{journal}\t{impact_factor}\n{journal_full}\t{impact_factor}\n'.format(**locals())
                        out.write(line)

                if payload['currentpage'] >= max_page:
                # if payload['currentpage'] >= 10:
                    self.logger.info('\033[32mAll {currentpage} page done!\033[0m'.format(**payload))
                    break
                    
                payload['currentpage'] += 1

        print journal_done

        utils.text2sqlite3(self.__dict__['database'], out_xls)
        

def main():

    g = GetIF(**args)

    if args['update']:
        g.update_database()

    elif args['searchname']:
        g.get_impact_factor()

    else:
        parser.print_help()


if __name__ == "__main__":

    parser = argparse.ArgumentParser(
        prog='getIF',
        version='%(prog)s {}'.format(__version__),
        usage='%(prog)s <searchname> [options]',
        description='\t\033[1mget impact factor for given journal name\033[0m',
        epilog='contact: {0} < {0}@novogene.com> '.format(__author__),
        formatter_class=argparse.RawTextHelpFormatter)

    parser.add_argument('searchname', help='the input journal name', nargs='*')
    parser.add_argument(
        '-d',
        '--database',
        help='the database name of impact factor[default=%(default)s]',
        default=os.path.join(BASE_DIR, 'impact_factor.sqlite3'))

    parser.add_argument('-u', '--update', help='update the database from website', action='store_true')

    args = vars(parser.parse_args())

    main()



#!/usr/bin/env python
# -*- coding=utf-8 -*-
import os
import time
import random
import sqlite3
import logging
import functools
import bs4
import requests


# 自定义的装饰器
def try_again(N=10, default='.'):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            n = 0
            while n < N:
                try:
                    return func(*args, **kwargs)
                except Exception as e:
                    n += 1
                    print '[warn] `{}` failed {}st time: {}'.format(func.__name__, n, e)
                    time.sleep(random.randint(5, 10))
            if n == N:
                print '[error] `{}` failed too many times({})'.format(func.__name__, N)
                return default
        return wrapper
    return decorator


@try_again()
def get_soup(url, **kwargs):

    headers = {
        'User-Agent':
        'Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/61.0.3163.100 Safari/537.36'
    }

    response = requests.get(url, headers=headers, **kwargs)
    soup = bs4.BeautifulSoup(response.content, 'lxml')

    return soup


def logger(log_format='[%(asctime)s %(levelname)s] %(message)s', log_level=logging.INFO):

    logging.basicConfig(level=log_level, format=log_format)

    logger = logging.getLogger(__name__)

    return logger


def text2sqlite3(database, textfile):

    if os.path.exists(database):
        enter = raw_input('database alread exists, enter "r" to remove it, "b" to backup it[default], "q" to quit:').lower()
        if enter == 'q':
            exit()
        elif enter == 'r':
            os.remove(database)
        elif enter == 'b':
	    os.system('mv {database} {database}.bak'.format(**locals()))
        else:
            exit(1)
    
    conn = sqlite3.connect(database)
    cursor = conn.cursor()

    create_table_sql = '''
        CREATE TABLE `factor` (
            journal VARCHAR(100) PRIMARY KEY,
            impact_factor VARCHAR(20)
        );
    '''

    cursor.execute(create_table_sql)

    data = {}
    with open(textfile) as f:
        for line in f:
            if line.startswith('#') or not line.strip():
                continue
            journal, impact_factor = line.strip().rsplit('\t', 1)
            data.update({journal: impact_factor})

    for context in data.iteritems():
        insert_sql = 'INSERT INTO `factor` VALUES("{}", "{}")'.format(*context)
        print insert_sql
        cursor.execute(insert_sql)

    conn.commit()

    cursor.close()
    conn.close()

    print 'database upated done!'


def sqlite3_execute(database, sql):

    conn = sqlite3.connect(database)
    cursor = conn.cursor()

    cursor.execute(sql)

    result = cursor.fetchall()

    cursor.close()
    conn.close()

    return result


if __name__ == '__main__':
    
    import sys
    if len(sys.argv) < 3:
	print 'usage: python %s <database> <textfile>' % sys.argv[0]
	exit()

    text2sqlite3(*sys.argv[1:])
    



#!/usr/bin/env python
# -*- coding=utf-8 -*-
import os
from setuptools import setup, find_packages

from pubmed2 import info

BASE_DIR = os.path.dirname(os.path.abspath(__file__))

setup(
    name='pubmed2',
    version=info.__version__,
    author=info.__author__,
    author_email=info.__author_email__,
    description='search pubmeds quickly',
    long_description=open(os.path.join(BASE_DIR, 'README.rst')).read(),
    url='https://github.com/suqingdong/pubmed2',
    license='BSD License',
    install_requires=[
        'bs4',
        'requests',
        'xlwt',
        'colorama',
        'django',
        'googletrans',
        'fuzzywuzzy',
        'python-Levenshtein',
    ],
    packages=find_packages(),
    include_package_data=True,
    scripts=['pubmed2/pubmed.py'],
    entry_points = {  
        'console_scripts': [  
            'pubmed = pubmed2.pubmed:main'  
        ]  
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Operating System :: OS Independent',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Programming Language :: Python :: Implementation',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Topic :: Software Development :: Libraries'
    ],
)